.include "options.inc"

.include "macros.inc"
.include "param-encodings.inc"

;-------------
; Declare default infrastructure

OP_CATEGORY regs, "Register Operations"
OP_CATEGORY flow, "Flow Control"
OP_CATEGORY math, "Arithmetic"
OP_CATDOC math, "All add and sub variations push a carry bit, inc and dec do not."
OP_CATEGORY bits, "Bitwise Operations"

ZPVAR rptr, regs, 1, "Pointer to the head of the register stack, which is also r0, and the lowest used byte in zp."
ZPVAR pptr, regs, 1, "Pointer to the prior register."
ZPVAR iptr, flow, 2, "Instruction pointer.  Always points to the beginning of an instruction, with (zp),y addressing reading the parameters."
ZPVAR cstack, math, 1, "Carry stack.  MSB is the current carry bit."

.segment "ZPTEMP": zeropage
zptemp: .res 3


.segment "ACHERON"
.include "op-utils.asm"


;----------------------
; Included ops and features

.include "ops-shift.asm"
.include "ops-addsub.asm"
.include "ops-muldiv.asm"
.include "ops-callret.asm"
.include "ops-misc.asm"

.if FEATURE__GLOBALS
.include "feature-globals.asm"
.endif

.if FEATURE__EXCEPTIONS
.include "feature-exceptions.asm"
.endif

.if FEATURE__TRAPS
.include "feature-traps.asm"
.endif






;----------------------
; Instruction dispatch,
; along with branch instructions that need to reach it,
; and select small instructions to space save on bcc vs jmp


OP flipc, none, bits, "Flip the state of the most recent carry bit, leaving it on the carry stack."
 lda cstack
 eor #$80
 sta cstack
 bcc mainLoop1

OP hibyte, none, bits, "rP := upper byte of rP."
 lda 1,x
 sta 0,x
OP lobyte, none, bits, "rP := lower byte of rP."
 lda #0
 sta 1,x
 bcc mainLoop1

OP bswap, none, bits, "Swap high and low bytes of rP."
 ldy 0,x
 lda 1,x
 sta 0,x
 sty 1,x
 bcc mainLoop1

PSEUDO set, "reg, imm", regs, "Becomes clr, set8, or set16.", {" .if (imm > 255) .or (imm < 0)",EOL,"   set16 reg, imm",EOL," .elseif (imm > 0)",EOL,"   set8 reg, imm",EOL," .else",EOL,"   clr reg",EOL," .endif"}

OPE set8, rdimm8, regs, "rD := imm8"
 lda (iptr),y
 sta 0,x
 lda #0
 sta 1,x
 bcc mainLoop2c

OP stmb, rard, regs, "memory(rA) := low byte of rD"
 decode_rdra
 sty pptr
 lda 0,y
 sta (0,x)
 bcc mainLoop2c

OPE clr, rd, regs, "rD := 0"
 lda #0
 sta 0,x
 sta 1,x
 ;bcc mainLoop1 ;passthrough



;-----------------------------------------------------
OPE with, rd, regs, "Explicitly assign rP."
OP noop, none, flow, "No operation."
mainLoop1:
 inc iptr
 bne mainLoop0
_incHi:
 inc iptr+1
 jmp mainLoop0
;-----------------------------------------------------



OP andp, imm16, bits, "rP := rP & imm16"
 lda 0,x
 and (iptr),y
 sta 0,x
 iny
 lda 1,x
 and (iptr),y
 sta 1,x
 bcc mainLoop3

OP orp, imm16, bits, "rP := rP | imm16"
 lda 0,x
 ora (iptr),y
 sta 0,x
 iny
 lda 1,x
 ora (iptr),y
 sta 1,x
 bcc mainLoop3

OP xorp, imm16, bits, "rP := rP ^ imm16"
 lda 0,x
 eor (iptr),y
 sta 0,x
 iny
 lda 1,x
 ora (iptr),y
 sta 1,x
 bcc mainLoop3

OP ldrptr, rdra, regs, "rD := current address of rA.  Note that task switching or otherwise swapping out zp can leave this pointer dangling."
 decode_rdra
 sty 0,x
 lda #0
 sta 1,x
 bcc mainLoop2

OP ldmb, rard, regs, "rD := memory(rA), unsigned byte"
 decode_rdra
 sty pptr
 lda (0,x)
 sta 0,y
 lda #0
 sta 1,y
 bcc mainLoop2c

OP copy, rdra, regs, "rD := rA"
 decode_rdra
 lda 0,y
 sta 0,x
 lda 1,y
 sta 1,x
 bcc mainLoop2c

OPE set16, rdimm16, regs, "rD := imm16"
 lda (iptr),y
 sta 0,x
 iny
 lda (iptr),y
 sta 1,x
; bcc mainLoop3 ;fall through

;-----------------------------------------------------
mainLoop3:
 sec
 .byte $a9
mainLoop2:
 clc
mainLoop2c:  ; c = carry already clear
 lda #2
mainLoopAc:
 adc iptr
 sta iptr
 bcs _incHi

mainLoop0:
 ldy #0                      ; Grab the next opcode byte
_instructionTrapSelfmod:
 lda (iptr),y                ; selfmods to 'beq _instructionTrapJump' when the instruction trap is enabled
_instructionTrapResume:
 iny

.if ALLOW_EMBEDDED_DISPATCH
 cmp #__firstEmbeddedOpcode
 bcs _embeddedDispatch
.endif

  ; Unembedded dispatch
  convertUnembeddedToDispatch  ; from COMPACT_UNEMBEDDED_OPCODES
  sta dispatch1
  ldx pptr
dispatch1 = *+1
  jmp (unembeddedDispatchTable)

.if FEATURE__TRAPS
_instructionTrapJump:
 jmp continueFromInstructionTrap  ; default vector just continues, and is overwritten by the registered trap handler
.endif

.if ALLOW_EMBEDDED_DISPATCH
_embeddedDispatch:
 tax           ; save byte for extracting the low nybble later
 and #$f0
 lsr
 lsr
 lsr           ; clears carry
 adc #<(embeddedDispatchTable - (__firstEmbeddedOpcode >> 3))
 sta dispatch2
 txa
 and #$0f      ; convert low nybble to a register pointer
 asl
 adc rptr      ; should clear carry.  Failure if rD wraps past $ff, which is a stack underflow
 tax
 stx pptr      ; update rP to point to this reg
dispatch2 = *+1
 jmp (embeddedDispatchTable)
.endif
;-----------------------------------------------------



; Branch instructions are located close to the dispatch to allow native branch tests

OP bc, rel8, flow, "Pop a carry bit, branch if it is set."
 asl cstack
 bcc mainLoop2c
 bcs ba

OP bnc, rel8, flow, "Pop a carry bit, branch if it is clear."
 asl cstack
 bcs mainLoop2
 bcc ba

OP bneg, rel8, flow, "Branch if rP is negative."
 lda 1,x
 bpl mainLoop2c
 bmi ba

OP bpos, rel8, flow, "Branch if rP is non-negative."
 lda 1,x
 bmi mainLoop2c
 bpl ba

OP bz, rel8, flow, "Branch if rP is zero."
 lda 0,x
 ora 1,x
 bne mainLoop2c
 beq ba

OP bnz, rel8, flow, "Branch if rP is non-zero."
 lda 0,x
 ora 1,x
 beq mainLoop2c
 ;bne ba           ; fall through to ba

OP ba, rel8, flow, "Always branch."
 ldx #0           ; .X = high byte of 16-bit delta
 lda (iptr),y     ; .A = signed low byte
 bpl :+
  dex
:clc
 adc iptr
 sta iptr
 txa
 adc iptr+1
 sta iptr+1
 jmp mainLoop0

OP decloop, rdimm4p1rel8, flow, "rD := rD - imm4 (1-16), branch until it wraps past zero.  Does not affect carry stack."
 decode_rdimm4
 tya              ; subtract
 eor #$ff
 adc 0,x
 sta 0,x
 ldy #2           ; needed for ba to read the right parameter byte
 bcs ba           ; branch if low byte didn't underflow
 lda 1,x          ; test to see if high byte will underflow
 beq :+
  dec 1,x         ; yes, dec and branch
  bcc ba          ; carry is known clear due to not taking bcs above
:dec 1,x
 jmp mainLoop3

PSEUDO case, "imm, rel8", flow, "Becomes case8 or case16", {" .if (imm >= 0) .and (imm <=255)",EOL,"   case8 imm, rel8",EOL," .else",EOL,"   case16 imm, rel8",EOL," .endif"}

OP case8, imm8rel8, flow, "Branch if rP = imm8."
 lda 1,x          ; fail if high byte of rP is nonzero
 bne :+
 lda (iptr),y     ; test low byte
 iny              ; prepare .Y for ba to read the rel8 byte
 cmp 0,x
 beq ba
:jmp mainLoop3

OP case16, imm16rel8, flow, "Branch if rP = imm16."
 lda (iptr),y     ; test low byte
 iny
 cmp 0,x
 bne mainLoopY
 lda (iptr),y     ; test high byte
 iny
 cmp 1,x
 beq ba
mainLoopY:
 clc
mainLoopYc:
 tya
 jmp mainLoopAc




























;------------------------
; Closeout of scopes

ZPVAR rstackTop, regs, 0, "The memory location right after the register stack.  This is the value of rptr when the register stack is empty."

.segment "UTABLE"
unembeddedDispatchTableEnd:
.segment "ETABLE"
embeddedDispatchTableEnd:
.segment "OPDESC_"
.byte ")",EOL


.export num__unembedded, num__embedded, num__free, __firstEmbeddedOpcode
num__unembedded = (unembeddedDispatchTableEnd - unembeddedDispatchTable)/2 * opcodesPerUnembeddedInstruction
num__embedded = (embeddedDispatchTableEnd - embeddedDispatchTable)/2
__firstEmbeddedOpcode = ((1 << 4) - num__embedded) << 4
num__free =  __firstEmbeddedOpcode - (num__unembedded * opcodesPerUnembeddedInstruction)

.segment "INCFILE_IMPORTS"
.byte EOL
.segment "INCFILE_OPCODES"
.byte EOL
.segment "INCFILE_INSTRUCTIONS"
.byte EOL
.segment "INCFILE_BOTTOM"
.byte EOL
.byte "; Number of embedded opcodes:   $", HEXBYTE(num__embedded), EOL
.byte "; Number of unembedded opcodes: $", HEXBYTE(num__unembedded), EOL
.byte "; Number of opcodes remaining:  $", HEXBYTE(num__free), EOL

.if num__free < 0
;.error .concat("Instruction set overflow.  ",.string(num__embedded)," embedded * 16 + ",.string(num__unembedded)," unembedded")
.endif

.segment "ACHERON"
